/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


#include <cmath>
#include "Config.h"

#include "wx/wx.h"

#include "OpenGL.h"

#include "Images/Drawable.h"
#include "Images/Image.h"
#include "Midi/Track.h"
#include "Editors/KeyboardEditor.h"
#include "GUI/GraphicalTrack.h"
#include "Midi/Sequence.h"
#include "Pickers/MagneticGrid.h"
#include "Pickers/VolumeSlider.h"
#include "Images/ImageProvider.h"
#include "GUI/GLPane.h"

#include "Editors/RelativeXCoord.h"

namespace AriaMaestosa {
	
// ***********************************************************************************************************************************************************
// **********************************************************    CONSTRUCTOR      ****************************************************************************
// ***********************************************************************************************************************************************************

KeyboardEditor::KeyboardEditor(Track* track) : Editor(track)
{
    
	
    dragging_on_track=false;
    clickedOnNote=false;
    lastClickedNote=-1;
	
    sb_position=0.5;
}

KeyboardEditor::~KeyboardEditor()
{
    
}

// ***********************************************************************************************************************************************************
// *********************************************************    MOUSE EXITED      ****************************************************************************
// ***********************************************************************************************************************************************************

void KeyboardEditor::mouseExited(int dragX_arg, int mousey_current, int XBeforeDrag_arg,int mousey_initial)
{
	
	RelativeXCoord<int> mousex_current(sequence);
    mousex_current.setValue(dragX_arg, WINDOW);
    
    RelativeXCoord<int> mousex_initial(sequence);
    mousex_initial.setValue(XBeforeDrag_arg, WINDOW);
	
	selectNotesInRect(mousex_current, mousey_current, mousex_initial, mousey_initial);
	selecting = false;
	dragging_on_track = false;
	
	sequence->glPane->render();
}

// ***********************************************************************************************************************************************************
// ***********************************************************    MOUSE DOWN      ****************************************************************************
// ***********************************************************************************************************************************************************

void KeyboardEditor::mouseDown(const int argx, const int y)
{

    Editor::mouseDown(argx, y);
    
    RelativeXCoord<int> x(sequence);
    x.setValue(argx, WINDOW);
    
    // check if user is dragging on this track
    dragging_on_track=false;
    if(y<getYEnd()-5 and y>getEditorYStart() and x.getRelativeTo(WINDOW) < getWidth() - 24)
	{
        dragging_on_track=true;  
    }
	
	clickedOnNote=false;
	
    // check if user clicked on a note
	if(dragging_on_track)
	{
	
		const int noteAmount = track->getNoteAmount();
		for(int n=0; n<noteAmount; n++)
		{
			const int x1=track->getNoteStartInPixels(n) - sequence->getXScrollInPixels();
			const int x2=track->getNoteEndInPixels(n) - sequence->getXScrollInPixels();
			const int y1=track->getNotePitchID(n)*10 + getEditorYStart() - getYScrollInPixels();
			
			if(x.getRelativeTo(EDITOR) >x1 and x.getRelativeTo(EDITOR) <x2 and  y>y1 and y<y1+12)
			{
				clickedOnNote=true;
				
				if(track->isNoteSelected(n) and !sequence->glPane->isSelectLessPressed())
				{
					// clicked on a selected note
					lastClickedNote=-1;
				}
				else lastClickedNote=n;
				
				track->playNote(n);
				
			}//end if
		}//next
		
	}
    
    // user clicked on a note. select it.
    if(clickedOnNote && lastClickedNote!=-1)
	{
        track->selectNote(ALL_NOTES, false);
        track->selectNote(lastClickedNote, true);
    }
	
	selecting = false;
    
}

// ***********************************************************************************************************************************************************
// ***********************************************************    MOUSE DRAG      ****************************************************************************
// ***********************************************************************************************************************************************************

void KeyboardEditor::mouseDrag(const int mousex_current, const int mousey_current, const int mousex_initial, const int mousey_initial)
{
	
    Editor::mouseDrag(mousex_current, mousey_current, mousex_initial, mousey_initial);
	
	// if user is did not click on note and moved the mouse vertically, then he is selecting
	if(!clickedOnNote and abs(mousey_current-mousey_initial)>10 and dragging_on_track) selecting=true;
			
	// check mouse is still within track boundaries. if not, terminate selection
	if(selecting)
	{
		if( !(mousey_current<getYEnd()-5 and mousey_current>getEditorYStart() and mousex_current < getWidth() - 24) )
		{
			mouseExited(mousex_current, mousey_current, mousex_initial, mousey_initial);
		}
	}
	
}


// ***********************************************************************************************************************************************************
// ***********************************************************    MOUSE UP      ******************************************************************************
// ***********************************************************************************************************************************************************

void KeyboardEditor::mouseUp(int dragX_arg, int mousey_current, int XBeforeDrag_arg, int mousey_initial)
{

    Editor::mouseUp(dragX_arg, mousey_current, XBeforeDrag_arg, mousey_initial);
    
    RelativeXCoord<int> mousex_current(sequence);
    mousex_current.setValue(dragX_arg, WINDOW);
    
    RelativeXCoord<int> mousex_initial(sequence);
    mousex_initial.setValue(XBeforeDrag_arg, WINDOW);
    
    if(dragging_on_track)
	{
        
        if(!clickedOnNote)
		{

			// --------------------------  select notes  ---------------------
			if(selecting)
			{
				selectNotesInRect(mousex_current, mousey_current, mousex_initial, mousey_initial);
            }
			else
			{
                // --------------------------  add note  --------------------------
                

                if(mousex_current.getRelativeTo(WINDOW) > sbBarDrawable->x or mousex_initial.getRelativeTo(WINDOW) > sbBarDrawable->x) return;

                
                int snapped_x1=snapMidiTickToGrid(mousex_initial.getRelativeTo(MIDI) + sequence->getXScrollInMidiTicks());
                int snapped_x2=snapMidiTickToGrid(mousex_current.getRelativeTo(MIDI) + sequence->getXScrollInMidiTicks());

                int note = (mousey_initial - getEditorYStart() + getYScrollInPixels())/10;
                
                track->addNote(note, snapped_x1, snapped_x2, 80 );
                    
            }
            
        }
        
        // ---------------------------  move note  -----------------------
        
        if(clickedOnNote)
		{
            
            int xstep=mousex_current.getRelativeTo(MIDI)-mousex_initial.getRelativeTo(MIDI);
            int ystep=mousey_current-mousey_initial;
            
            xstep=snapMidiTickToGrid(xstep);
            ystep=(int)round(ystep/10.0);
            
			if(xstep == 0 and ystep == 0)
			{
				goto end_of_func;
			}
			
            // move a single note
            if(lastClickedNote!=-1)
                track->moveNotes(xstep, ystep, lastClickedNote);
            else
                track->moveNotes(xstep, ystep, SELECTED_NOTES);
            
        }// end if clicked on note
        
    }//end if dragging_on_track
    
    
end_of_func:
		
    dragging_on_track=false;
    clickedOnNote=false;
}

// --------------  select notes --------
void KeyboardEditor::selectNotesInRect(RelativeXCoord<int>& mousex_current, int mousey_current,
									   RelativeXCoord<int>& mousex_initial, int mousey_initial)
{
	for(int n=0; n<track->getNoteAmount(); n++)
	{
		
		int x1=track->getNoteStartInPixels(n);
		int x2=track->getNoteEndInPixels(n);
		int from_y=track->getNotePitchID(n);
		
		if( x1>std::min( mousex_current.getRelativeTo(EDITOR), mousex_initial.getRelativeTo(EDITOR) ) + sequence->getXScrollInPixels() and
			x2<std::max( mousex_current.getRelativeTo(EDITOR), mousex_initial.getRelativeTo(EDITOR) ) + sequence->getXScrollInPixels() and
			from_y*10+getEditorYStart()-getYScrollInPixels() > std::min(mousey_current, mousey_initial) and
			from_y*10+getEditorYStart()-getYScrollInPixels() < std::max(mousey_current, mousey_initial) )
		{
			
			track->selectNote(n, true);
			
		}else{
			track->selectNote(n, false);
		}
	}//next
	
}

void KeyboardEditor::mouseHeldDown()
{
    Editor::mouseHeldDown();
}

int KeyboardEditor::getYScrollInPixels()
{
    return (int)(   sb_position*(120*11-height-20)   );
}

// ***********************************************************************************************************************************************************
// ************************************************************    RENDER      *******************************************************************************
// ***********************************************************************************************************************************************************

void KeyboardEditor::render(int dragX_arg, int mousey_current, int XBeforeDrag_arg, int mousey_initial, bool focus)
{
    
    if(!imagesLoaded()) return;

    assert(sbArrowDrawable->image!=NULL);
    assert(sbBackgDrawable->image!=NULL);
    assert(sbBarDrawable->image!=NULL);
    assert(noteTrackDrawable->image!=NULL);
    assert(noteTrackBackgDrawable->image!=NULL);

    // prepare coords
    RelativeXCoord<int> mousex_current(sequence);
    mousex_current.setValue(dragX_arg, WINDOW);
    
    RelativeXCoord<int> mousex_initial(sequence);
    mousex_initial.setValue(XBeforeDrag_arg, WINDOW);
    
    glEnable(GL_SCISSOR_TEST);
    // glScissor doesn't seem to follow the coordinate system so this ends up in all kinds of weird code to map to my coord system (from_y going down)
    glScissor(10, sequence->glPane->getHeight() - (20+height + from_y+barHeight+20), width-15, 20+height);
    
    // ------------------ draw lined background ----------------
    glEnable(GL_TEXTURE_2D);
    
    for(int g_octaveID=0; g_octaveID<11; g_octaveID++){
        int g_octave_y=g_octaveID*120-getYScrollInPixels();
        if(g_octave_y>-120 and g_octave_y<height+20){
            
            glColor4f(1,1,1,1);
            
            noteTrackBackgDrawable->move(64, from_y+barHeight+21 + g_octave_y);
            noteTrackBackgDrawable->scale( (width - 5 - 65)/noteTrackBackgDrawable->image->width ,1);    
            noteTrackBackgDrawable->render();
            
        }//end if
    }//next
    
    
    drawVerticalMeasureLines(getEditorYStart(), getYEnd());

    // ---------------------- draw notes ----------------------------
    const int noteAmount = track->getNoteAmount();
    for(int n=0; n<noteAmount; n++)
    {

        int x1=track->getNoteStartInPixels(n) - sequence->getXScrollInPixels();
        int x2=track->getNoteEndInPixels(n) - sequence->getXScrollInPixels();
        
        // don't draw notes that won't visible
        if(x2<0) continue;
        if(x1>width) break;
        
        int y=track->getNotePitchID(n);
        float volume=track->getNoteVolume(n)/127.0;
        
        if(track->isNoteSelected(n) and focus) glColor3f((1-volume)*1, (1-(volume/2))*1, 0);
        else glColor3f((1-volume)*0.9, (1-volume)*0.9, (1-volume)*0.9);
        
        glBegin(GL_QUADS);
        glVertex2f(x1+getEditorXStart()+1, y*10+1 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x1+getEditorXStart()+1, y*10+10 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x2+getEditorXStart()-1, y*10+10 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x2+getEditorXStart()-1, y*10+1 + getEditorYStart() - getYScrollInPixels());
        glEnd();

// opengl lines displayed at wrong pixel on my Linux system, tehrefore i wrote this hack
#ifdef __WXGTK__
        glColor3f(0, 0, 0);
        glBegin(GL_QUADS);
        glVertex2f(x1+getEditorXStart()+1, y*10+1 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x1+getEditorXStart()+1, y*10+10 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x1+getEditorXStart(), y*10+10 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x1+getEditorXStart(), y*10+1 + getEditorYStart() - getYScrollInPixels());

        glVertex2f(x1+getEditorXStart()+1, y*10+10 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x2+getEditorXStart()-1, y*10+10 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x2+getEditorXStart()-1, y*10+11 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x1+getEditorXStart()+1, y*10+11 + getEditorYStart() - getYScrollInPixels());

        glVertex2f(x2+getEditorXStart(), y*10+10 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x2+getEditorXStart(), y*10+1 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x2+getEditorXStart()-1, y*10+1 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x2+getEditorXStart()-1, y*10+10 + getEditorYStart() - getYScrollInPixels());

        glVertex2f(x1+getEditorXStart()+1, y*10+1 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x2+getEditorXStart()-1, y*10+1 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x2+getEditorXStart()-1, y*10 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x1+getEditorXStart()+1, y*10 + getEditorYStart() - getYScrollInPixels());
        glEnd();

#else
        glColor3f(0, 0, 0);
        glBegin(GL_LINES);
        glVertex2f(x1+getEditorXStart()+1, y*10+1 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x1+getEditorXStart()+1, y*10+10 + getEditorYStart() - getYScrollInPixels());
        
        glVertex2f(x1+getEditorXStart()+1, y*10+11 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x2+getEditorXStart()-1, y*10+11 + getEditorYStart() - getYScrollInPixels());
        
        glVertex2f(x2+getEditorXStart(), y*10+10 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x2+getEditorXStart(), y*10+1 + getEditorYStart() - getYScrollInPixels());
        
        glVertex2f(x1+getEditorXStart()+1, y*10+1 + getEditorYStart() - getYScrollInPixels());
        glVertex2f(x2+getEditorXStart()-1, y*10+1 + getEditorYStart() - getYScrollInPixels());
        glEnd();
#endif
        
    }
    
    
    // ------------------ draw keyboard ----------------
    
    // grey background
    glDisable(GL_TEXTURE_2D);
    if(!focus) glColor3f(0.4, 0.4, 0.4);
    else glColor3f(0.8, 0.8, 0.8);
    glBegin(GL_QUADS);
    
    glVertex2f( 0, getEditorYStart());
    glVertex2f( 0, getYEnd());
    glVertex2f( getEditorXStart()-25, getYEnd());
    glVertex2f( getEditorXStart()-25, getEditorYStart());
    
    glEnd();
    
    for(int g_octaveID=0; g_octaveID<11; g_octaveID++)
    {
        int g_octave_y=g_octaveID*120-getYScrollInPixels();
        if(g_octave_y>-120 and g_octave_y<height+20)
        {
            
            glEnable(GL_TEXTURE_2D);
            
            if(!focus) glColor3f(0.5, 0.5, 0.5);
            else glColor3f(1,1,1);
            
            noteTrackDrawable->move(getEditorXStart()-noteTrackDrawable->image->width, from_y+barHeight+20 + g_octave_y);
            noteTrackDrawable->render();
            
            
            glLoadIdentity();
            glDisable(GL_TEXTURE_2D);
            glColor3f(0,0,0);
            
            glBegin(GL_LINES);
            glVertex2f(0, from_y+barHeight+20 + g_octave_y+1);
            glVertex2f(getEditorXStart()-25, from_y+barHeight+20 + g_octave_y+1);
            glEnd();
            
            glRasterPos2f(30,from_y+barHeight+21 + g_octave_y +120/2);
            
            char buffer[2];
            sprintf (buffer, "%d", 10-g_octaveID);
            
            for(int i=0; buffer[i]; i++){
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, buffer[i]);
            }
            glEnable(GL_TEXTURE_2D);
            
        }//end if
    }//next

    
    
    glEnable(GL_TEXTURE_2D);
    
    // ------------------------- mouse drag (preview) ------------------------
    
    if(!clickedOnNote)
    {
        if(dragging_on_track)
        {
            
            // selection
            //if(abs(mousey_current-mousey_initial)>10)
            //{
			if(selecting)
			{
                glDisable(GL_TEXTURE_2D);
                glColor3f(0,0,0);
                glBegin(GL_LINES);
                
                glVertex2f(mousex_initial.getRelativeTo(WINDOW), mousey_current);
                glVertex2f(mousex_initial.getRelativeTo(WINDOW), mousey_initial);
                
                glVertex2f(mousex_current.getRelativeTo(WINDOW), mousey_initial);
                glVertex2f(mousex_initial.getRelativeTo(WINDOW), mousey_initial);
                
                glVertex2f(mousex_current.getRelativeTo(WINDOW), mousey_initial);
                glVertex2f(mousex_current.getRelativeTo(WINDOW), mousey_current);
                
                glVertex2f(mousex_initial.getRelativeTo(WINDOW), mousey_current);
                glVertex2f(mousex_current.getRelativeTo(WINDOW), mousey_current);
                
                glEnd();
                glEnable(GL_TEXTURE_2D);
                
            }
			else
			{
                // ----------------------- add note (preview) --------------------
                
                    glDisable(GL_TEXTURE_2D);
                    glColor3f(1, 0.85, 0);
                    
                    
                    int preview_x1=
                        (int)(
                              (snapMidiTickToGrid(mousex_initial.getRelativeTo(MIDI) + sequence->getXScrollInMidiTicks()) -
                               sequence->getXScrollInMidiTicks())*sequence->getZoom()
                              );
                    int preview_x2=
                        (int)(
                              (snapMidiTickToGrid(mousex_current.getRelativeTo(MIDI) + sequence->getXScrollInMidiTicks()) -
                               sequence->getXScrollInMidiTicks())*sequence->getZoom()
                              );
                    
                    if(!(preview_x1<0 || preview_x2<0)){
                        
                    glBegin(GL_QUADS);
                    
                    glVertex2f(preview_x1+getEditorXStart(), ((mousey_initial - getEditorYStart() + getYScrollInPixels())/10)*10 + getEditorYStart() - getYScrollInPixels());
                    glVertex2f(preview_x1+getEditorXStart(), ((mousey_initial - getEditorYStart() + getYScrollInPixels())/10)*10+10 + getEditorYStart() - getYScrollInPixels());
                    glVertex2f(preview_x2+getEditorXStart(), ((mousey_initial - getEditorYStart() + getYScrollInPixels())/10)*10+10 + getEditorYStart() - getYScrollInPixels());
                    glVertex2f(preview_x2+getEditorXStart(), ((mousey_initial - getEditorYStart() + getYScrollInPixels())/10)*10 + getEditorYStart() - getYScrollInPixels());
                    
                    glEnd();

                    }
                    
                    glEnable(GL_TEXTURE_2D);
            }// end if selection or addition
        }// end if dragging on track
        
    } // end if !clickedOnNote
    
    // ------------------------- move note (preview) -----------------------
    if(clickedOnNote)
    {
        
        glDisable(GL_TEXTURE_2D);
        
        glColor4f(1, 0.85, 0, 0.5);
        
        int xstep=mousex_current.getRelativeTo(MIDI)-mousex_initial.getRelativeTo(MIDI);
        int ystep=mousey_current-mousey_initial;
        
        xstep = (int)( snapMidiTickToGrid(xstep) * sequence->getZoom() );
        ystep=(int)round(ystep/10.0);
            
        // move a single note
        if(lastClickedNote!=-1)
        {
            int x1=track->getNoteStartInPixels(lastClickedNote) - sequence->getXScrollInPixels();
            int x2=track->getNoteEndInPixels(lastClickedNote) - sequence->getXScrollInPixels();
            int y=track->getNotePitchID(lastClickedNote);
            
            glBegin(GL_QUADS);
            glVertex2f(x1+xstep+getEditorXStart(), (y+ystep)*10+1 + getEditorYStart() - getYScrollInPixels());
            glVertex2f(x1+xstep+getEditorXStart(), (y+ystep)*10+10 + getEditorYStart() - getYScrollInPixels());
            glVertex2f(x2-1+xstep+getEditorXStart(), (y+ystep)*10+10 + getEditorYStart() - getYScrollInPixels());
            glVertex2f(x2-1+xstep+getEditorXStart(), (y+ystep)*10+1 + getEditorYStart() - getYScrollInPixels());
            glEnd();
            
        }
        else
        {
            // move a bunch of notes
            
            for(int n=0; n<track->getNoteAmount(); n++)
            {
                if(!track->isNoteSelected(n)) continue;
                
                int x1=track->getNoteStartInPixels(n) - sequence->getXScrollInPixels();
                int x2=track->getNoteEndInPixels(n) - sequence->getXScrollInPixels();
                int y=track->getNotePitchID(n);
                //float volume=track->getNoteVolume(n)/127.0;
                
                glBegin(GL_QUADS);
                glVertex2f(x1+xstep+getEditorXStart(), (y+ystep)*10+1 + getEditorYStart() - getYScrollInPixels());
                glVertex2f(x1+xstep+getEditorXStart(), (y+ystep)*10+10 + getEditorYStart() - getYScrollInPixels());
                glVertex2f(x2-1+xstep+getEditorXStart(), (y+ystep)*10+10 + getEditorYStart() - getYScrollInPixels());
                glVertex2f(x2-1+xstep+getEditorXStart(), (y+ystep)*10+1 + getEditorYStart() - getYScrollInPixels());
                glEnd();
            }//next
            
        }
        
        glEnable(GL_TEXTURE_2D);
        glLoadIdentity();
        
    }

    // ---------------------------- scrollbar -----------------------
    if(!focus) glColor3f(0.5, 0.5, 0.5);
    else glColor3f(1,1,1);
    
    renderScrollbar();
    
    glColor3f(1,1,1);
    
    glDisable(GL_SCISSOR_TEST);
    
}


void KeyboardEditor::rightClick(const int x, const int y)
{
    
    // iterate through notes and find which one the user clicked on (if any)
    for(int n=0; n<track->getNoteAmount(); n++){
        int x1=track->getNoteStartInPixels(n) - sequence->getXScrollInPixels()+getEditorXStart();
        int x2=track->getNoteEndInPixels(n) - sequence->getXScrollInPixels()+getEditorXStart();
        int y1=track->getNotePitchID(n)*10 + getEditorYStart() - getYScrollInPixels();

        if(x>x1 and x<x2 and  y>y1 and y<y1+10){
            
            wxPoint winCoord=sequence->glPane->ClientToScreen(wxPoint(x,y));
            volumeSlider->show(winCoord.x, winCoord.y, n, track);
            
            return;
            
        }// end if mouse within note bounds
        
    }// next note
        
}

}
